package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.ClassFilePool;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.env.AccessRestriction;
import org.eclipse.jdt.internal.compiler.env.IBinaryAnnotation;
import org.eclipse.jdt.internal.compiler.env.IBinaryType;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.env.INameEnvironmentExtension;
import org.eclipse.jdt.internal.compiler.env.ITypeAnnotationWalker;
import org.eclipse.jdt.internal.compiler.env.NameEnvironmentAnswer;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ITypeRequestor;
import org.eclipse.jdt.internal.compiler.lookup.AnnotatableTypeSystem;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.BaseTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.BinaryTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.CaptureBinding;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.ElementValuePair;
import org.eclipse.jdt.internal.compiler.lookup.ExternalAnnotationSuperimposer;
import org.eclipse.jdt.internal.compiler.lookup.IQualifiedTypeResolutionListener;
import org.eclipse.jdt.internal.compiler.lookup.ImportBinding;
import org.eclipse.jdt.internal.compiler.lookup.InferenceContext18;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier15;
import org.eclipse.jdt.internal.compiler.lookup.MissingTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.PolyTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.PolymorphicMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemPackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemReasons;
import org.eclipse.jdt.internal.compiler.lookup.ProblemReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.RawTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SignatureWrapper;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeSystem;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.UnresolvedAnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.UnresolvedReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.VoidTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.WildcardBinding;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.HashtableOfPackage;
import org.eclipse.jdt.internal.compiler.util.SimpleLookupTable;

public class LookupEnvironment implements ProblemReasons, TypeConstants {
  private Map accessRestrictions;
  
  ImportBinding[] defaultImports;
  
  public PackageBinding defaultPackage;
  
  HashtableOfPackage knownPackages;
  
  private int lastCompletedUnitIndex = -1;
  
  private int lastUnitIndex = -1;
  
  TypeSystem typeSystem;
  
  public INameEnvironment nameEnvironment;
  
  public CompilerOptions globalOptions;
  
  public ProblemReporter problemReporter;
  
  public ClassFilePool classFilePool;
  
  private int stepCompleted;
  
  public ITypeRequestor typeRequestor;
  
  private SimpleLookupTable uniqueParameterizedGenericMethodBindings;
  
  private SimpleLookupTable uniquePolymorphicMethodBindings;
  
  private SimpleLookupTable uniqueGetClassMethodBinding;
  
  public CompilationUnitDeclaration unitBeingCompleted = null;
  
  public Object missingClassFileLocation = null;
  
  private CompilationUnitDeclaration[] units = new CompilationUnitDeclaration[4];
  
  private MethodVerifier verifier;
  
  public MethodBinding arrayClone;
  
  private ArrayList missingTypes;
  
  Set<SourceTypeBinding> typesBeingConnected;
  
  public boolean isProcessingAnnotations = false;
  
  public boolean mayTolerateMissingType = false;
  
  PackageBinding nullableAnnotationPackage;
  
  PackageBinding nonnullAnnotationPackage;
  
  PackageBinding nonnullByDefaultAnnotationPackage;
  
  AnnotationBinding nonNullAnnotation;
  
  AnnotationBinding nullableAnnotation;
  
  Map<String, Integer> allNullAnnotations = null;
  
  final List<MethodBinding> deferredEnumMethods = new ArrayList<>();
  
  InferenceContext18 currentInferenceContext;
  
  static final int BUILD_FIELDS_AND_METHODS = 4;
  
  static final int BUILD_TYPE_HIERARCHY = 1;
  
  static final int CHECK_AND_SET_IMPORTS = 2;
  
  static final int CONNECT_TYPE_HIERARCHY = 3;
  
  static final ProblemPackageBinding TheNotFoundPackage = new ProblemPackageBinding(CharOperation.NO_CHAR, 1);
  
  static final ProblemReferenceBinding TheNotFoundType = new ProblemReferenceBinding(CharOperation.NO_CHAR_CHAR, null, 1);
  
  public IQualifiedTypeResolutionListener[] resolutionListeners;
  
  public ReferenceBinding askForType(char[][] compoundName) {
    NameEnvironmentAnswer answer = this.nameEnvironment.findType(compoundName);
    if (answer == null)
      return null; 
    if (answer.isBinaryType()) {
      this.typeRequestor.accept(answer.getBinaryType(), computePackageFrom(compoundName, false), answer.getAccessRestriction());
    } else if (answer.isCompilationUnit()) {
      this.typeRequestor.accept(answer.getCompilationUnit(), answer.getAccessRestriction());
    } else if (answer.isSourceType()) {
      this.typeRequestor.accept(answer.getSourceTypes(), computePackageFrom(compoundName, false), answer.getAccessRestriction());
    } 
    return getCachedType(compoundName);
  }
  
  ReferenceBinding askForType(PackageBinding packageBinding, char[] name) {
    if (packageBinding == null)
      packageBinding = this.defaultPackage; 
    NameEnvironmentAnswer answer = this.nameEnvironment.findType(name, packageBinding.compoundName);
    if (answer == null)
      return null; 
    if (answer.isBinaryType()) {
      this.typeRequestor.accept(answer.getBinaryType(), packageBinding, answer.getAccessRestriction());
    } else if (answer.isCompilationUnit()) {
      try {
        this.typeRequestor.accept(answer.getCompilationUnit(), answer.getAccessRestriction());
      } catch (AbortCompilation abort) {
        if (CharOperation.equals(name, TypeConstants.PACKAGE_INFO_NAME))
          return null; 
        throw abort;
      } 
    } else if (answer.isSourceType()) {
      this.typeRequestor.accept(answer.getSourceTypes(), packageBinding, answer.getAccessRestriction());
      ReferenceBinding binding = packageBinding.getType0(name);
      String externalAnnotationPath = answer.getExternalAnnotationPath();
      if (externalAnnotationPath != null && this.globalOptions.isAnnotationBasedNullAnalysisEnabled && binding instanceof SourceTypeBinding)
        ExternalAnnotationSuperimposer.apply((SourceTypeBinding)binding, externalAnnotationPath); 
      return binding;
    } 
    return packageBinding.getType0(name);
  }
  
  public void buildTypeBindings(CompilationUnitDeclaration unit, AccessRestriction accessRestriction) {
    CompilationUnitScope scope = new CompilationUnitScope(unit, this);
    scope.buildTypeBindings(accessRestriction);
    int unitsLength = this.units.length;
    if (++this.lastUnitIndex >= unitsLength)
      System.arraycopy(this.units, 0, this.units = new CompilationUnitDeclaration[2 * unitsLength], 0, unitsLength); 
    this.units[this.lastUnitIndex] = unit;
  }
  
  public BinaryTypeBinding cacheBinaryType(IBinaryType binaryType, AccessRestriction accessRestriction) {
    return cacheBinaryType(binaryType, true, accessRestriction);
  }
  
  public BinaryTypeBinding cacheBinaryType(IBinaryType binaryType, boolean needFieldsAndMethods, AccessRestriction accessRestriction) {
    char[][] compoundName = CharOperation.splitOn('/', binaryType.getName());
    ReferenceBinding existingType = getCachedType(compoundName);
    if (existingType == null || existingType instanceof UnresolvedReferenceBinding)
      return createBinaryTypeFrom(binaryType, computePackageFrom(compoundName, false), needFieldsAndMethods, accessRestriction); 
    return null;
  }
  
  public void completeTypeBindings() {
    this.stepCompleted = 1;
    int i;
    for (i = this.lastCompletedUnitIndex + 1; i <= this.lastUnitIndex; i++)
      (this.unitBeingCompleted = this.units[i]).scope.checkAndSetImports(); 
    this.stepCompleted = 2;
    for (i = this.lastCompletedUnitIndex + 1; i <= this.lastUnitIndex; i++)
      (this.unitBeingCompleted = this.units[i]).scope.connectTypeHierarchy(); 
    this.stepCompleted = 3;
    for (i = this.lastCompletedUnitIndex + 1; i <= this.lastUnitIndex; i++) {
      CompilationUnitScope unitScope = (this.unitBeingCompleted = this.units[i]).scope;
      unitScope.checkParameterizedTypes();
      unitScope.buildFieldsAndMethods();
      this.units[i] = null;
    } 
    this.stepCompleted = 4;
    this.lastCompletedUnitIndex = this.lastUnitIndex;
    this.unitBeingCompleted = null;
  }
  
  public void completeTypeBindings(CompilationUnitDeclaration parsedUnit) {
    if (this.stepCompleted == 4) {
      completeTypeBindings();
    } else {
      if (parsedUnit.scope == null)
        return; 
      if (this.stepCompleted >= 2)
        (this.unitBeingCompleted = parsedUnit).scope.checkAndSetImports(); 
      if (this.stepCompleted >= 3)
        (this.unitBeingCompleted = parsedUnit).scope.connectTypeHierarchy(); 
      this.unitBeingCompleted = null;
    } 
  }
  
  public void completeTypeBindings(CompilationUnitDeclaration parsedUnit, boolean buildFieldsAndMethods) {
    if (parsedUnit.scope == null)
      return; 
    (this.unitBeingCompleted = parsedUnit).scope.checkAndSetImports();
    parsedUnit.scope.connectTypeHierarchy();
    parsedUnit.scope.checkParameterizedTypes();
    if (buildFieldsAndMethods)
      parsedUnit.scope.buildFieldsAndMethods(); 
    this.unitBeingCompleted = null;
  }
  
  public void completeTypeBindings(CompilationUnitDeclaration[] parsedUnits, boolean[] buildFieldsAndMethods, int unitCount) {
    int i;
    for (i = 0; i < unitCount; i++) {
      CompilationUnitDeclaration parsedUnit = parsedUnits[i];
      if (parsedUnit.scope != null)
        (this.unitBeingCompleted = parsedUnit).scope.checkAndSetImports(); 
    } 
    for (i = 0; i < unitCount; i++) {
      CompilationUnitDeclaration parsedUnit = parsedUnits[i];
      if (parsedUnit.scope != null)
        (this.unitBeingCompleted = parsedUnit).scope.connectTypeHierarchy(); 
    } 
    for (i = 0; i < unitCount; i++) {
      CompilationUnitDeclaration parsedUnit = parsedUnits[i];
      if (parsedUnit.scope != null) {
        (this.unitBeingCompleted = parsedUnit).scope.checkParameterizedTypes();
        if (buildFieldsAndMethods[i])
          parsedUnit.scope.buildFieldsAndMethods(); 
      } 
    } 
    this.unitBeingCompleted = null;
  }
  
  public MethodBinding computeArrayClone(MethodBinding objectClone) {
    if (this.arrayClone == null)
      this.arrayClone = new MethodBinding(
          objectClone.modifiers & 0xFFFFFFFB | 0x1, 
          TypeConstants.CLONE, 
          objectClone.returnType, 
          Binding.NO_PARAMETERS, 
          Binding.NO_EXCEPTIONS, 
          (ReferenceBinding)objectClone.returnType); 
    return this.arrayClone;
  }
  
  public TypeBinding computeBoxingType(TypeBinding type) {
    ReferenceBinding referenceBinding8;
    ReferenceBinding referenceBinding7;
    ReferenceBinding referenceBinding6;
    ReferenceBinding referenceBinding5;
    ReferenceBinding referenceBinding4;
    ReferenceBinding referenceBinding3;
    ReferenceBinding referenceBinding2;
    ReferenceBinding referenceBinding1;
    switch (type.id) {
      case 33:
        return (TypeBinding)TypeBinding.BOOLEAN;
      case 26:
        return (TypeBinding)TypeBinding.BYTE;
      case 28:
        return (TypeBinding)TypeBinding.CHAR;
      case 27:
        return (TypeBinding)TypeBinding.SHORT;
      case 32:
        return (TypeBinding)TypeBinding.DOUBLE;
      case 31:
        return (TypeBinding)TypeBinding.FLOAT;
      case 29:
        return (TypeBinding)TypeBinding.INT;
      case 30:
        return (TypeBinding)TypeBinding.LONG;
      case 10:
        referenceBinding8 = getType(JAVA_LANG_INTEGER);
        if (referenceBinding8 != null)
          return (TypeBinding)referenceBinding8; 
        return (TypeBinding)new ProblemReferenceBinding(JAVA_LANG_INTEGER, null, 1);
      case 3:
        referenceBinding7 = getType(JAVA_LANG_BYTE);
        if (referenceBinding7 != null)
          return (TypeBinding)referenceBinding7; 
        return (TypeBinding)new ProblemReferenceBinding(JAVA_LANG_BYTE, null, 1);
      case 4:
        referenceBinding6 = getType(JAVA_LANG_SHORT);
        if (referenceBinding6 != null)
          return (TypeBinding)referenceBinding6; 
        return (TypeBinding)new ProblemReferenceBinding(JAVA_LANG_SHORT, null, 1);
      case 2:
        referenceBinding5 = getType(JAVA_LANG_CHARACTER);
        if (referenceBinding5 != null)
          return (TypeBinding)referenceBinding5; 
        return (TypeBinding)new ProblemReferenceBinding(JAVA_LANG_CHARACTER, null, 1);
      case 7:
        referenceBinding4 = getType(JAVA_LANG_LONG);
        if (referenceBinding4 != null)
          return (TypeBinding)referenceBinding4; 
        return (TypeBinding)new ProblemReferenceBinding(JAVA_LANG_LONG, null, 1);
      case 9:
        referenceBinding3 = getType(JAVA_LANG_FLOAT);
        if (referenceBinding3 != null)
          return (TypeBinding)referenceBinding3; 
        return (TypeBinding)new ProblemReferenceBinding(JAVA_LANG_FLOAT, null, 1);
      case 8:
        referenceBinding2 = getType(JAVA_LANG_DOUBLE);
        if (referenceBinding2 != null)
          return (TypeBinding)referenceBinding2; 
        return (TypeBinding)new ProblemReferenceBinding(JAVA_LANG_DOUBLE, null, 1);
      case 5:
        referenceBinding1 = getType(JAVA_LANG_BOOLEAN);
        if (referenceBinding1 != null)
          return (TypeBinding)referenceBinding1; 
        return (TypeBinding)new ProblemReferenceBinding(JAVA_LANG_BOOLEAN, null, 1);
    } 
    switch (type.kind()) {
      case 516:
      case 4100:
      case 8196:
        switch ((type.erasure()).id) {
          case 33:
            return (TypeBinding)TypeBinding.BOOLEAN;
          case 26:
            return (TypeBinding)TypeBinding.BYTE;
          case 28:
            return (TypeBinding)TypeBinding.CHAR;
          case 27:
            return (TypeBinding)TypeBinding.SHORT;
          case 32:
            return (TypeBinding)TypeBinding.DOUBLE;
          case 31:
            return (TypeBinding)TypeBinding.FLOAT;
          case 29:
            return (TypeBinding)TypeBinding.INT;
          case 30:
            return (TypeBinding)TypeBinding.LONG;
        } 
        break;
      case 65540:
        return ((PolyTypeBinding)type).computeBoxingType();
      case 32772:
        return computeBoxingType((TypeBinding)type.getIntersectingTypes()[0]);
    } 
    return type;
  }
  
  private PackageBinding computePackageFrom(char[][] constantPoolName, boolean isMissing) {
    if (constantPoolName.length == 1)
      return this.defaultPackage; 
    PackageBinding packageBinding = getPackage0(constantPoolName[0]);
    if (packageBinding == null || packageBinding == TheNotFoundPackage) {
      packageBinding = new PackageBinding(constantPoolName[0], this);
      if (isMissing)
        packageBinding.tagBits |= 0x80L; 
      this.knownPackages.put(constantPoolName[0], packageBinding);
    } 
    for (int i = 1, length = constantPoolName.length - 1; i < length; i++) {
      PackageBinding parent = packageBinding;
      if ((packageBinding = parent.getPackage0(constantPoolName[i])) == null || packageBinding == TheNotFoundPackage) {
        packageBinding = new PackageBinding(CharOperation.subarray(constantPoolName, 0, i + 1), parent, this);
        if (isMissing)
          packageBinding.tagBits |= 0x80L; 
        parent.addPackage(packageBinding);
      } 
    } 
    return packageBinding;
  }
  
  public ReferenceBinding convertToParameterizedType(ReferenceBinding originalType) {
    if (originalType != null) {
      boolean isGeneric = originalType.isGenericType();
      ReferenceBinding originalEnclosingType = originalType.enclosingType();
      ReferenceBinding convertedEnclosingType = originalEnclosingType;
      boolean needToConvert = isGeneric;
      if (originalEnclosingType != null) {
        convertedEnclosingType = originalType.isStatic() ? 
          (ReferenceBinding)convertToRawType((TypeBinding)originalEnclosingType, false) : 
          convertToParameterizedType(originalEnclosingType);
        needToConvert |= TypeBinding.notEquals((TypeBinding)originalEnclosingType, (TypeBinding)convertedEnclosingType);
      } 
      if (needToConvert)
        return (ReferenceBinding)createParameterizedType(originalType, isGeneric ? (TypeBinding[])originalType.typeVariables() : null, convertedEnclosingType); 
    } 
    return originalType;
  }
  
  public TypeBinding convertToRawType(TypeBinding type, boolean forceRawEnclosingType) {
    int dimension;
    TypeBinding originalType;
    boolean needToConvert;
    ParameterizedTypeBinding paramType;
    TypeBinding convertedType;
    switch (type.kind()) {
      case 132:
      case 516:
      case 1028:
      case 4100:
      case 8196:
        return type;
      case 68:
        dimension = type.dimensions();
        originalType = type.leafComponentType();
        break;
      default:
        if (type.id == 1)
          return type; 
        dimension = 0;
        originalType = type;
        break;
    } 
    switch (originalType.kind()) {
      case 132:
        return type;
      case 2052:
        needToConvert = true;
        break;
      case 260:
        paramType = (ParameterizedTypeBinding)originalType;
        needToConvert = paramType.genericType().isGenericType();
        break;
      default:
        needToConvert = false;
        break;
    } 
    ReferenceBinding originalEnclosing = originalType.enclosingType();
    if (originalEnclosing == null) {
      convertedType = needToConvert ? (TypeBinding)createRawType((ReferenceBinding)originalType.erasure(), null) : originalType;
    } else {
      int i;
      boolean bool;
      ReferenceBinding convertedEnclosing;
      if (originalEnclosing.kind() == 1028) {
        i = needToConvert | (((ReferenceBinding)originalType).isStatic() ? 0 : 1);
        convertedEnclosing = originalEnclosing;
      } else if (forceRawEnclosingType && i == 0) {
        convertedEnclosing = (ReferenceBinding)convertToRawType((TypeBinding)originalEnclosing, forceRawEnclosingType);
        bool = TypeBinding.notEquals((TypeBinding)originalEnclosing, (TypeBinding)convertedEnclosing);
      } else if (bool || ((ReferenceBinding)originalType).isStatic()) {
        convertedEnclosing = (ReferenceBinding)convertToRawType((TypeBinding)originalEnclosing, false);
      } else {
        convertedEnclosing = convertToParameterizedType(originalEnclosing);
      } 
      if (bool) {
        RawTypeBinding rawTypeBinding = createRawType((ReferenceBinding)originalType.erasure(), convertedEnclosing);
      } else if (TypeBinding.notEquals((TypeBinding)originalEnclosing, (TypeBinding)convertedEnclosing)) {
        ParameterizedTypeBinding parameterizedTypeBinding = createParameterizedType((ReferenceBinding)originalType.erasure(), null, convertedEnclosing);
      } else {
        convertedType = originalType;
      } 
    } 
    if (TypeBinding.notEquals(originalType, convertedType))
      return (dimension > 0) ? (TypeBinding)createArrayType(convertedType, dimension) : convertedType; 
    return type;
  }
  
  public ReferenceBinding[] convertToRawTypes(ReferenceBinding[] originalTypes, boolean forceErasure, boolean forceRawEnclosingType) {
    if (originalTypes == null)
      return null; 
    ReferenceBinding[] convertedTypes = originalTypes;
    for (int i = 0, length = originalTypes.length; i < length; i++) {
      ReferenceBinding originalType = originalTypes[i];
      ReferenceBinding convertedType = (ReferenceBinding)convertToRawType(forceErasure ? originalType.erasure() : (TypeBinding)originalType, forceRawEnclosingType);
      if (TypeBinding.notEquals((TypeBinding)convertedType, (TypeBinding)originalType)) {
        if (convertedTypes == originalTypes)
          System.arraycopy(originalTypes, 0, convertedTypes = new ReferenceBinding[length], 0, i); 
        convertedTypes[i] = convertedType;
      } else if (convertedTypes != originalTypes) {
        convertedTypes[i] = originalType;
      } 
    } 
    return convertedTypes;
  }
  
  public TypeBinding convertUnresolvedBinaryToRawType(TypeBinding type) {
    int dimension;
    TypeBinding originalType;
    boolean needToConvert;
    ParameterizedTypeBinding paramType;
    TypeBinding convertedType;
    switch (type.kind()) {
      case 132:
      case 516:
      case 1028:
      case 4100:
      case 8196:
        return type;
      case 68:
        dimension = type.dimensions();
        originalType = type.leafComponentType();
        break;
      default:
        if (type.id == 1)
          return type; 
        dimension = 0;
        originalType = type;
        break;
    } 
    switch (originalType.kind()) {
      case 132:
        return type;
      case 2052:
        needToConvert = true;
        break;
      case 260:
        paramType = (ParameterizedTypeBinding)originalType;
        needToConvert = paramType.genericType().isGenericType();
        break;
      default:
        needToConvert = false;
        break;
    } 
    ReferenceBinding originalEnclosing = originalType.enclosingType();
    if (originalEnclosing == null) {
      convertedType = needToConvert ? (TypeBinding)createRawType((ReferenceBinding)originalType.erasure(), null) : originalType;
    } else {
      int i;
      ReferenceBinding convertedEnclosing = (ReferenceBinding)convertUnresolvedBinaryToRawType((TypeBinding)originalEnclosing);
      if (TypeBinding.notEquals((TypeBinding)convertedEnclosing, (TypeBinding)originalEnclosing))
        i = needToConvert | (((ReferenceBinding)originalType).isStatic() ? 0 : 1); 
      if (i != 0) {
        RawTypeBinding rawTypeBinding = createRawType((ReferenceBinding)originalType.erasure(), convertedEnclosing);
      } else if (TypeBinding.notEquals((TypeBinding)originalEnclosing, (TypeBinding)convertedEnclosing)) {
        ParameterizedTypeBinding parameterizedTypeBinding = createParameterizedType((ReferenceBinding)originalType.erasure(), null, convertedEnclosing);
      } else {
        convertedType = originalType;
      } 
    } 
    if (TypeBinding.notEquals(originalType, convertedType))
      return (dimension > 0) ? (TypeBinding)createArrayType(convertedType, dimension) : convertedType; 
    return type;
  }
  
  public AnnotationBinding createAnnotation(ReferenceBinding annotationType, ElementValuePair[] pairs) {
    if (pairs.length != 0) {
      AnnotationBinding.setMethodBindings(annotationType, pairs);
      return new AnnotationBinding(annotationType, pairs);
    } 
    return this.typeSystem.getAnnotationType(annotationType, true);
  }
  
  public AnnotationBinding createUnresolvedAnnotation(ReferenceBinding annotationType, ElementValuePair[] pairs) {
    if (pairs.length != 0)
      return (AnnotationBinding)new UnresolvedAnnotationBinding(annotationType, pairs, this); 
    return this.typeSystem.getAnnotationType(annotationType, false);
  }
  
  public ArrayBinding createArrayType(TypeBinding leafComponentType, int dimensionCount) {
    return this.typeSystem.getArrayType(leafComponentType, dimensionCount);
  }
  
  public ArrayBinding createArrayType(TypeBinding leafComponentType, int dimensionCount, AnnotationBinding[] annotations) {
    return this.typeSystem.getArrayType(leafComponentType, dimensionCount, annotations);
  }
  
  public TypeBinding createIntersectionType18(ReferenceBinding[] intersectingTypes) {
    return this.typeSystem.getIntersectionType18(intersectingTypes);
  }
  
  public BinaryTypeBinding createBinaryTypeFrom(IBinaryType binaryType, PackageBinding packageBinding, AccessRestriction accessRestriction) {
    return createBinaryTypeFrom(binaryType, packageBinding, true, accessRestriction);
  }
  
  public BinaryTypeBinding createBinaryTypeFrom(IBinaryType binaryType, PackageBinding packageBinding, boolean needFieldsAndMethods, AccessRestriction accessRestriction) {
    BinaryTypeBinding binaryBinding = new BinaryTypeBinding(packageBinding, binaryType, this);
    ReferenceBinding cachedType = packageBinding.getType0(binaryBinding.compoundName[binaryBinding.compoundName.length - 1]);
    if (cachedType != null && !cachedType.isUnresolvedType()) {
      if (cachedType.isBinaryBinding())
        return (BinaryTypeBinding)cachedType; 
      return null;
    } 
    packageBinding.addType((ReferenceBinding)binaryBinding);
    setAccessRestriction((ReferenceBinding)binaryBinding, accessRestriction);
    binaryBinding.cachePartsFrom(binaryType, needFieldsAndMethods);
    return binaryBinding;
  }
  
  public MissingTypeBinding createMissingType(PackageBinding packageBinding, char[][] compoundName) {
    if (packageBinding == null) {
      packageBinding = computePackageFrom(compoundName, true);
      if (packageBinding == TheNotFoundPackage)
        packageBinding = this.defaultPackage; 
    } 
    MissingTypeBinding missingType = new MissingTypeBinding(packageBinding, compoundName, this);
    if (missingType.id != 1) {
      MissingTypeBinding missingTypeBinding;
      ReferenceBinding objectType = getType(TypeConstants.JAVA_LANG_OBJECT);
      if (objectType == null)
        missingTypeBinding = createMissingType(null, TypeConstants.JAVA_LANG_OBJECT); 
      missingType.setMissingSuperclass((ReferenceBinding)missingTypeBinding);
    } 
    packageBinding.addType((ReferenceBinding)missingType);
    if (this.missingTypes == null)
      this.missingTypes = new ArrayList(3); 
    this.missingTypes.add(missingType);
    return missingType;
  }
  
  public PackageBinding createPackage(char[][] compoundName) {
    PackageBinding packageBinding = getPackage0(compoundName[0]);
    if (packageBinding == null || packageBinding == TheNotFoundPackage) {
      packageBinding = new PackageBinding(compoundName[0], this);
      this.knownPackages.put(compoundName[0], packageBinding);
    } 
    for (int i = 1, length = compoundName.length; i < length; i++) {
      ReferenceBinding type = packageBinding.getType0(compoundName[i]);
      if (type != null && type != TheNotFoundType && !(type instanceof UnresolvedReferenceBinding))
        return null; 
      PackageBinding parent = packageBinding;
      if ((packageBinding = parent.getPackage0(compoundName[i])) == null || packageBinding == TheNotFoundPackage) {
        if (this.nameEnvironment instanceof INameEnvironmentExtension) {
          if (((INameEnvironmentExtension)this.nameEnvironment).findType(compoundName[i], parent.compoundName, false) != null)
            return null; 
        } else if (this.nameEnvironment.findType(compoundName[i], parent.compoundName) != null) {
          return null;
        } 
        packageBinding = new PackageBinding(CharOperation.subarray(compoundName, 0, i + 1), parent, this);
        parent.addPackage(packageBinding);
      } 
    } 
    return packageBinding;
  }
  
  public ParameterizedGenericMethodBinding createParameterizedGenericMethod(MethodBinding genericMethod, RawTypeBinding rawType) {
    ParameterizedGenericMethodBinding[] cachedInfo = (ParameterizedGenericMethodBinding[])this.uniqueParameterizedGenericMethodBindings.get(genericMethod);
    boolean needToGrow = false;
    int index = 0;
    if (cachedInfo != null) {
      for (int max = cachedInfo.length; index < max; ) {
        ParameterizedGenericMethodBinding cachedMethod = cachedInfo[index];
        if (cachedMethod == null)
          break; 
        if (!cachedMethod.isRaw || 
          cachedMethod.declaringClass != ((rawType == null) ? genericMethod.declaringClass : (ReferenceBinding)rawType)) {
          index++;
          continue;
        } 
        return cachedMethod;
      } 
      needToGrow = true;
    } else {
      cachedInfo = new ParameterizedGenericMethodBinding[5];
      this.uniqueParameterizedGenericMethodBindings.put(genericMethod, cachedInfo);
    } 
    int length = cachedInfo.length;
    if (needToGrow && index == length) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new ParameterizedGenericMethodBinding[length * 2], 0, length);
      this.uniqueParameterizedGenericMethodBindings.put(genericMethod, cachedInfo);
    } 
    ParameterizedGenericMethodBinding parameterizedGenericMethod = new ParameterizedGenericMethodBinding(genericMethod, rawType, this);
    cachedInfo[index] = parameterizedGenericMethod;
    return parameterizedGenericMethod;
  }
  
  public ParameterizedGenericMethodBinding createParameterizedGenericMethod(MethodBinding genericMethod, TypeBinding[] typeArguments) {
    return createParameterizedGenericMethod(genericMethod, typeArguments, false, false);
  }
  
  public ParameterizedGenericMethodBinding createParameterizedGenericMethod(MethodBinding genericMethod, TypeBinding[] typeArguments, boolean inferredWithUncheckedConversion, boolean hasReturnProblem) {
    ParameterizedGenericMethodBinding[] cachedInfo = (ParameterizedGenericMethodBinding[])this.uniqueParameterizedGenericMethodBindings.get(genericMethod);
    int argLength = (typeArguments == null) ? 0 : typeArguments.length;
    boolean needToGrow = false;
    int index = 0;
    if (cachedInfo != null) {
      for (int max = cachedInfo.length; index < max; index++) {
        ParameterizedGenericMethodBinding cachedMethod = cachedInfo[index];
        if (cachedMethod == null)
          break; 
        if (cachedMethod.isRaw || 
          cachedMethod.inferredWithUncheckedConversion != inferredWithUncheckedConversion)
          continue; 
        TypeBinding[] cachedArguments = cachedMethod.typeArguments;
        int cachedArgLength = (cachedArguments == null) ? 0 : cachedArguments.length;
        if (argLength != cachedArgLength)
          continue; 
        for (int j = 0; j < cachedArgLength; j++) {
          if (typeArguments[j] != cachedArguments[j])
            // Byte code: goto -> 229 
        } 
        if (inferredWithUncheckedConversion) {
          if (cachedMethod.returnType.isParameterizedType() || cachedMethod.returnType.isTypeVariable())
            continue; 
          byte b;
          int i;
          ReferenceBinding[] arrayOfReferenceBinding;
          for (i = (arrayOfReferenceBinding = cachedMethod.thrownExceptions).length, b = 0; b < i; ) {
            ReferenceBinding referenceBinding = arrayOfReferenceBinding[b];
            if (!referenceBinding.isParameterizedType()) {
              if (referenceBinding.isTypeVariable())
                // Byte code: goto -> 229 
              b++;
            } 
            // Byte code: goto -> 229
          } 
        } 
        return cachedMethod;
      } 
      needToGrow = true;
    } else {
      cachedInfo = new ParameterizedGenericMethodBinding[5];
      this.uniqueParameterizedGenericMethodBindings.put(genericMethod, cachedInfo);
    } 
    int length = cachedInfo.length;
    if (needToGrow && index == length) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new ParameterizedGenericMethodBinding[length * 2], 0, length);
      this.uniqueParameterizedGenericMethodBindings.put(genericMethod, cachedInfo);
    } 
    ParameterizedGenericMethodBinding parameterizedGenericMethod = 
      new ParameterizedGenericMethodBinding(genericMethod, typeArguments, this, inferredWithUncheckedConversion, hasReturnProblem);
    cachedInfo[index] = parameterizedGenericMethod;
    return parameterizedGenericMethod;
  }
  
  public PolymorphicMethodBinding createPolymorphicMethod(MethodBinding originalPolymorphicMethod, TypeBinding[] parameters) {
    String key = new String(originalPolymorphicMethod.selector);
    PolymorphicMethodBinding[] cachedInfo = (PolymorphicMethodBinding[])this.uniquePolymorphicMethodBindings.get(key);
    int parametersLength = (parameters == null) ? 0 : parameters.length;
    TypeBinding[] parametersTypeBinding = new TypeBinding[parametersLength];
    for (int i = 0; i < parametersLength; i++) {
      TypeBinding parameterTypeBinding = parameters[i];
      if (parameterTypeBinding.id == 12) {
        parametersTypeBinding[i] = (TypeBinding)getType(JAVA_LANG_VOID);
      } else {
        parametersTypeBinding[i] = parameterTypeBinding.erasure();
      } 
    } 
    boolean needToGrow = false;
    int index = 0;
    if (cachedInfo != null) {
      for (int max = cachedInfo.length; index < max; index++) {
        PolymorphicMethodBinding cachedMethod = cachedInfo[index];
        if (cachedMethod == null)
          break; 
        if (cachedMethod.matches(parametersTypeBinding, originalPolymorphicMethod.returnType))
          return cachedMethod; 
      } 
      needToGrow = true;
    } else {
      cachedInfo = new PolymorphicMethodBinding[5];
      this.uniquePolymorphicMethodBindings.put(key, cachedInfo);
    } 
    int length = cachedInfo.length;
    if (needToGrow && index == length) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new PolymorphicMethodBinding[length * 2], 0, length);
      this.uniquePolymorphicMethodBindings.put(key, cachedInfo);
    } 
    PolymorphicMethodBinding polymorphicMethod = new PolymorphicMethodBinding(
        originalPolymorphicMethod, 
        parametersTypeBinding);
    cachedInfo[index] = polymorphicMethod;
    return polymorphicMethod;
  }
  
  public boolean usesAnnotatedTypeSystem() {
    return this.typeSystem.isAnnotatedTypeSystem();
  }
  
  public MethodBinding updatePolymorphicMethodReturnType(PolymorphicMethodBinding binding, TypeBinding typeBinding) {
    String key = new String(binding.selector);
    PolymorphicMethodBinding[] cachedInfo = (PolymorphicMethodBinding[])this.uniquePolymorphicMethodBindings.get(key);
    boolean needToGrow = false;
    int index = 0;
    TypeBinding[] parameters = binding.parameters;
    if (cachedInfo != null) {
      for (int max = cachedInfo.length; index < max; index++) {
        PolymorphicMethodBinding cachedMethod = cachedInfo[index];
        if (cachedMethod == null)
          break; 
        if (cachedMethod.matches(parameters, typeBinding))
          return (MethodBinding)cachedMethod; 
      } 
      needToGrow = true;
    } else {
      cachedInfo = new PolymorphicMethodBinding[5];
      this.uniquePolymorphicMethodBindings.put(key, cachedInfo);
    } 
    int length = cachedInfo.length;
    if (needToGrow && index == length) {
      System.arraycopy(cachedInfo, 0, cachedInfo = new PolymorphicMethodBinding[length * 2], 0, length);
      this.uniquePolymorphicMethodBindings.put(key, cachedInfo);
    } 
    PolymorphicMethodBinding polymorphicMethod = new PolymorphicMethodBinding(
        binding.original(), 
        typeBinding, 
        parameters);
    cachedInfo[index] = polymorphicMethod;
    return (MethodBinding)polymorphicMethod;
  }
  
  public ParameterizedMethodBinding createGetClassMethod(TypeBinding receiverType, MethodBinding originalMethod, Scope scope) {
    ParameterizedMethodBinding retVal = null;
    if (this.uniqueGetClassMethodBinding == null) {
      this.uniqueGetClassMethodBinding = new SimpleLookupTable(3);
    } else {
      retVal = (ParameterizedMethodBinding)this.uniqueGetClassMethodBinding.get(receiverType);
    } 
    if (retVal == null) {
      retVal = ParameterizedMethodBinding.instantiateGetClass(receiverType, originalMethod, scope);
      this.uniqueGetClassMethodBinding.put(receiverType, retVal);
    } 
    return retVal;
  }
  
  public ReferenceBinding createMemberType(ReferenceBinding memberType, ReferenceBinding enclosingType) {
    return this.typeSystem.getMemberType(memberType, enclosingType);
  }
  
  public ParameterizedTypeBinding createParameterizedType(ReferenceBinding genericType, TypeBinding[] typeArguments, ReferenceBinding enclosingType) {
    AnnotationBinding[] annotations = genericType.typeAnnotations;
    if (annotations != Binding.NO_ANNOTATIONS)
      return this.typeSystem.getParameterizedType((ReferenceBinding)genericType.unannotated(), typeArguments, enclosingType, annotations); 
    return this.typeSystem.getParameterizedType(genericType, typeArguments, enclosingType);
  }
  
  public ParameterizedTypeBinding createParameterizedType(ReferenceBinding genericType, TypeBinding[] typeArguments, ReferenceBinding enclosingType, AnnotationBinding[] annotations) {
    return this.typeSystem.getParameterizedType(genericType, typeArguments, enclosingType, annotations);
  }
  
  public TypeBinding createAnnotatedType(TypeBinding type, AnnotationBinding[][] annotations) {
    return this.typeSystem.getAnnotatedType(type, annotations);
  }
  
  public TypeBinding createAnnotatedType(TypeBinding type, AnnotationBinding[] newbies) {
    int newLength = (newbies == null) ? 0 : newbies.length;
    if (type == null || newLength == 0)
      return type; 
    AnnotationBinding[] oldies = type.getTypeAnnotations();
    int oldLength = (oldies == null) ? 0 : oldies.length;
    if (oldLength > 0) {
      System.arraycopy(newbies, 0, newbies = new AnnotationBinding[newLength + oldLength], 0, newLength);
      System.arraycopy(oldies, 0, newbies, newLength, oldLength);
    } 
    if (this.globalOptions.isAnnotationBasedNullAnalysisEnabled) {
      long tagBitsSeen = 0L;
      AnnotationBinding[] filtered = new AnnotationBinding[newbies.length];
      int count = 0;
      for (int i = 0; i < newbies.length; i++) {
        if (newbies[i] == null) {
          filtered[count++] = null;
        } else {
          long tagBits = 0L;
          if ((newbies[i]).type.hasNullBit(32)) {
            tagBits = 72057594037927936L;
          } else if ((newbies[i]).type.hasNullBit(64)) {
            tagBits = 36028797018963968L;
          } 
          if ((tagBitsSeen & tagBits) == 0L) {
            tagBitsSeen |= tagBits;
            filtered[count++] = newbies[i];
          } 
        } 
      } 
      if (count < newbies.length)
        System.arraycopy(filtered, 0, newbies = new AnnotationBinding[count], 0, count); 
    } 
    return this.typeSystem.getAnnotatedType(type, new AnnotationBinding[][] { newbies });
  }
  
  public RawTypeBinding createRawType(ReferenceBinding genericType, ReferenceBinding enclosingType) {
    AnnotationBinding[] annotations = genericType.typeAnnotations;
    if (annotations != Binding.NO_ANNOTATIONS)
      return this.typeSystem.getRawType((ReferenceBinding)genericType.unannotated(), enclosingType, annotations); 
    return this.typeSystem.getRawType(genericType, enclosingType);
  }
  
  public RawTypeBinding createRawType(ReferenceBinding genericType, ReferenceBinding enclosingType, AnnotationBinding[] annotations) {
    return this.typeSystem.getRawType(genericType, enclosingType, annotations);
  }
  
  public WildcardBinding createWildcard(ReferenceBinding genericType, int rank, TypeBinding bound, TypeBinding[] otherBounds, int boundKind) {
    if (genericType != null) {
      AnnotationBinding[] annotations = genericType.typeAnnotations;
      if (annotations != Binding.NO_ANNOTATIONS)
        return this.typeSystem.getWildcard((ReferenceBinding)genericType.unannotated(), rank, bound, otherBounds, boundKind, annotations); 
    } 
    return this.typeSystem.getWildcard(genericType, rank, bound, otherBounds, boundKind);
  }
  
  public CaptureBinding createCapturedWildcard(WildcardBinding wildcard, ReferenceBinding contextType, int start, int end, ASTNode cud, int id) {
    return this.typeSystem.getCapturedWildcard(wildcard, contextType, start, end, cud, id);
  }
  
  public WildcardBinding createWildcard(ReferenceBinding genericType, int rank, TypeBinding bound, TypeBinding[] otherBounds, int boundKind, AnnotationBinding[] annotations) {
    return this.typeSystem.getWildcard(genericType, rank, bound, otherBounds, boundKind, annotations);
  }
  
  public AccessRestriction getAccessRestriction(TypeBinding type) {
    return (AccessRestriction)this.accessRestrictions.get(type);
  }
  
  public ReferenceBinding getCachedType(char[][] compoundName) {
    if (compoundName.length == 1)
      return this.defaultPackage.getType0(compoundName[0]); 
    PackageBinding packageBinding = getPackage0(compoundName[0]);
    if (packageBinding == null || packageBinding == TheNotFoundPackage)
      return null; 
    for (int i = 1, packageLength = compoundName.length - 1; i < packageLength; i++) {
      if ((packageBinding = packageBinding.getPackage0(compoundName[i])) == null || packageBinding == TheNotFoundPackage)
        return null; 
    } 
    return packageBinding.getType0(compoundName[compoundName.length - 1]);
  }
  
  public AnnotationBinding getNullableAnnotation() {
    if (this.nullableAnnotation != null)
      return this.nullableAnnotation; 
    ReferenceBinding nullable = getResolvedType(this.globalOptions.nullableAnnotationName, null);
    return this.nullableAnnotation = this.typeSystem.getAnnotationType(nullable, true);
  }
  
  public char[][] getNullableAnnotationName() {
    return this.globalOptions.nullableAnnotationName;
  }
  
  public AnnotationBinding getNonNullAnnotation() {
    if (this.nonNullAnnotation != null)
      return this.nonNullAnnotation; 
    ReferenceBinding nonNull = getResolvedType(this.globalOptions.nonNullAnnotationName, null);
    return this.nonNullAnnotation = this.typeSystem.getAnnotationType(nonNull, true);
  }
  
  public AnnotationBinding[] nullAnnotationsFromTagBits(long nullTagBits) {
    if (nullTagBits == 72057594037927936L)
      return new AnnotationBinding[] { getNonNullAnnotation() }; 
    if (nullTagBits == 36028797018963968L)
      return new AnnotationBinding[] { getNullableAnnotation() }; 
    return null;
  }
  
  public char[][] getNonNullAnnotationName() {
    return this.globalOptions.nonNullAnnotationName;
  }
  
  public char[][] getNonNullByDefaultAnnotationName() {
    return this.globalOptions.nonNullByDefaultAnnotationName;
  }
  
  int getNullAnnotationBit(char[][] qualifiedTypeName) {
    if (this.allNullAnnotations == null) {
      this.allNullAnnotations = new HashMap<>();
      this.allNullAnnotations.put(CharOperation.toString(this.globalOptions.nonNullAnnotationName), Integer.valueOf(32));
      this.allNullAnnotations.put(CharOperation.toString(this.globalOptions.nullableAnnotationName), Integer.valueOf(64));
      this.allNullAnnotations.put(CharOperation.toString(this.globalOptions.nonNullByDefaultAnnotationName), Integer.valueOf(128));
      byte b;
      int i;
      String[] arrayOfString;
      for (i = (arrayOfString = this.globalOptions.nullableAnnotationSecondaryNames).length, b = 0; b < i; ) {
        String name = arrayOfString[b];
        this.allNullAnnotations.put(name, Integer.valueOf(64));
        b++;
      } 
      for (i = (arrayOfString = this.globalOptions.nonNullAnnotationSecondaryNames).length, b = 0; b < i; ) {
        String name = arrayOfString[b];
        this.allNullAnnotations.put(name, Integer.valueOf(32));
        b++;
      } 
      for (i = (arrayOfString = this.globalOptions.nonNullByDefaultAnnotationSecondaryNames).length, b = 0; b < i; ) {
        String name = arrayOfString[b];
        this.allNullAnnotations.put(name, Integer.valueOf(128));
        b++;
      } 
    } 
    String qualifiedTypeString = CharOperation.toString(qualifiedTypeName);
    Integer typeBit = this.allNullAnnotations.get(qualifiedTypeString);
    return (typeBit == null) ? 0 : typeBit.intValue();
  }
  
  public boolean isNullnessAnnotationPackage(PackageBinding pkg) {
    return !(this.nonnullAnnotationPackage != pkg && this.nullableAnnotationPackage != pkg && this.nonnullByDefaultAnnotationPackage != pkg);
  }
  
  public boolean usesNullTypeAnnotations() {
    if (this.globalOptions.useNullTypeAnnotations != null)
      return this.globalOptions.useNullTypeAnnotations.booleanValue(); 
    initializeUsesNullTypeAnnotation();
    for (MethodBinding enumMethod : this.deferredEnumMethods) {
      int purpose = 0;
      if (CharOperation.equals(enumMethod.selector, TypeConstants.VALUEOF)) {
        purpose = 10;
      } else if (CharOperation.equals(enumMethod.selector, TypeConstants.VALUES)) {
        purpose = 9;
      } 
      if (purpose != 0)
        SyntheticMethodBinding.markNonNull(enumMethod, purpose, this); 
    } 
    this.deferredEnumMethods.clear();
    return this.globalOptions.useNullTypeAnnotations.booleanValue();
  }
  
  private void initializeUsesNullTypeAnnotation() {
    this.globalOptions.useNullTypeAnnotations = Boolean.FALSE;
    if (!this.globalOptions.isAnnotationBasedNullAnalysisEnabled || this.globalOptions.originalSourceLevel < 3407872L)
      return; 
    ReferenceBinding nullable = (this.nullableAnnotation != null) ? this.nullableAnnotation.getAnnotationType() : getType(getNullableAnnotationName());
    ReferenceBinding nonNull = (this.nonNullAnnotation != null) ? this.nonNullAnnotation.getAnnotationType() : getType(getNonNullAnnotationName());
    if (nullable == null && nonNull == null)
      return; 
    if (nullable == null || nonNull == null)
      return; 
    long nullableMetaBits = nullable.getAnnotationTagBits() & 0x20000000000000L;
    long nonNullMetaBits = nonNull.getAnnotationTagBits() & 0x20000000000000L;
    if (nullableMetaBits != nonNullMetaBits)
      return; 
    if (nullableMetaBits == 0L)
      return; 
    this.globalOptions.useNullTypeAnnotations = Boolean.TRUE;
  }
  
  PackageBinding getPackage0(char[] name) {
    return this.knownPackages.get(name);
  }
  
  public ReferenceBinding getResolvedType(char[][] compoundName, Scope scope) {
    ReferenceBinding type = getType(compoundName);
    if (type != null)
      return type; 
    this.problemReporter.isClassPathCorrect(
        compoundName, 
        (scope == null) ? this.unitBeingCompleted : scope.referenceCompilationUnit(), 
        this.missingClassFileLocation);
    return (ReferenceBinding)createMissingType(null, compoundName);
  }
  
  PackageBinding getTopLevelPackage(char[] name) {
    PackageBinding packageBinding = getPackage0(name);
    if (packageBinding != null) {
      if (packageBinding == TheNotFoundPackage)
        return null; 
      return packageBinding;
    } 
    if (this.nameEnvironment.isPackage(null, name)) {
      this.knownPackages.put(name, packageBinding = new PackageBinding(name, this));
      return packageBinding;
    } 
    this.knownPackages.put(name, (PackageBinding)TheNotFoundPackage);
    return null;
  }
  
  public ReferenceBinding getType(char[][] compoundName) {
    if (compoundName.length == 1) {
      if ((referenceBinding = this.defaultPackage.getType0(compoundName[0])) == null) {
        PackageBinding packageBinding = getPackage0(compoundName[0]);
        if (packageBinding != null && packageBinding != TheNotFoundPackage)
          return null; 
        referenceBinding = askForType(this.defaultPackage, compoundName[0]);
      } 
    } else {
      PackageBinding packageBinding = getPackage0(compoundName[0]);
      if (packageBinding == TheNotFoundPackage)
        return null; 
      if (packageBinding != null)
        for (int i = 1, packageLength = compoundName.length - 1; i < packageLength && (
          packageBinding = packageBinding.getPackage0(compoundName[i])) != null; i++) {
          if (packageBinding == TheNotFoundPackage)
            return null; 
        }  
      if (packageBinding == null) {
        referenceBinding = askForType(compoundName);
      } else if ((referenceBinding = packageBinding.getType0(compoundName[compoundName.length - 1])) == null) {
        referenceBinding = askForType(packageBinding, compoundName[compoundName.length - 1]);
      } 
    } 
    if (referenceBinding == null || referenceBinding == TheNotFoundType)
      return null; 
    ReferenceBinding referenceBinding = (ReferenceBinding)BinaryTypeBinding.resolveType((TypeBinding)referenceBinding, this, false);
    if (referenceBinding.isNestedType())
      return (ReferenceBinding)new ProblemReferenceBinding(compoundName, referenceBinding, 4); 
    return referenceBinding;
  }
  
  private TypeBinding[] getTypeArgumentsFromSignature(SignatureWrapper wrapper, TypeVariableBinding[] staticVariables, ReferenceBinding enclosingType, ReferenceBinding genericType, char[][][] missingTypeNames, ITypeAnnotationWalker walker) {
    ArrayList<TypeBinding> args = new ArrayList(2);
    int rank = 0;
    while (true) {
      args.add(getTypeFromVariantTypeSignature(wrapper, staticVariables, enclosingType, genericType, rank, missingTypeNames, 
            walker.toTypeArgument(rank++)));
      if (wrapper.signature[wrapper.start] == '>') {
        wrapper.start++;
        TypeBinding[] typeArguments = new TypeBinding[args.size()];
        args.toArray(typeArguments);
        return typeArguments;
      } 
    } 
  }
  
  private ReferenceBinding getTypeFromCompoundName(char[][] compoundName, boolean isParameterized, boolean wasMissingType) {
    UnresolvedReferenceBinding unresolvedReferenceBinding;
    ReferenceBinding referenceBinding1, binding = getCachedType(compoundName);
    if (binding == null) {
      PackageBinding packageBinding = computePackageFrom(compoundName, false);
      unresolvedReferenceBinding = new UnresolvedReferenceBinding(compoundName, packageBinding);
      if (wasMissingType)
        ((ReferenceBinding)unresolvedReferenceBinding).tagBits |= 0x80L; 
      packageBinding.addType((ReferenceBinding)unresolvedReferenceBinding);
    } else {
      MissingTypeBinding missingTypeBinding;
      if (unresolvedReferenceBinding == TheNotFoundType) {
        if (!wasMissingType)
          this.problemReporter.isClassPathCorrect(compoundName, this.unitBeingCompleted, this.missingClassFileLocation); 
        missingTypeBinding = createMissingType(null, compoundName);
      } else if (!isParameterized) {
        referenceBinding1 = (ReferenceBinding)convertUnresolvedBinaryToRawType((TypeBinding)missingTypeBinding);
      } 
    } 
    return referenceBinding1;
  }
  
  ReferenceBinding getTypeFromConstantPoolName(char[] signature, int start, int end, boolean isParameterized, char[][][] missingTypeNames, ITypeAnnotationWalker walker) {
    if (end == -1)
      end = signature.length; 
    char[][] compoundName = CharOperation.splitOn('/', signature, start, end);
    boolean wasMissingType = false;
    if (missingTypeNames != null)
      for (int i = 0, max = missingTypeNames.length; i < max; i++) {
        if (CharOperation.equals(compoundName, missingTypeNames[i])) {
          wasMissingType = true;
          break;
        } 
      }  
    ReferenceBinding binding = getTypeFromCompoundName(compoundName, isParameterized, wasMissingType);
    if (walker != ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER)
      binding = (ReferenceBinding)annotateType((TypeBinding)binding, walker, missingTypeNames); 
    return binding;
  }
  
  ReferenceBinding getTypeFromConstantPoolName(char[] signature, int start, int end, boolean isParameterized, char[][][] missingTypeNames) {
    return getTypeFromConstantPoolName(signature, start, end, isParameterized, missingTypeNames, ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER);
  }
  
  TypeBinding getTypeFromSignature(char[] signature, int start, int end, boolean isParameterized, TypeBinding enclosingType, char[][][] missingTypeNames, ITypeAnnotationWalker walker) {
    ReferenceBinding referenceBinding;
    TypeBinding typeBinding1;
    ArrayBinding arrayBinding;
    int dimension = 0;
    while (signature[start] == '[') {
      start++;
      dimension++;
    } 
    AnnotationBinding[][] annotationsOnDimensions = null;
    if (dimension > 0 && walker != ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER)
      for (int i = 0; i < dimension; i++) {
        AnnotationBinding[] annotations = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(0), this, missingTypeNames);
        if (annotations != Binding.NO_ANNOTATIONS) {
          if (annotationsOnDimensions == null)
            annotationsOnDimensions = new AnnotationBinding[dimension][]; 
          annotationsOnDimensions[i] = annotations;
        } 
        walker = walker.toNextArrayDimension();
      }  
    if (end == -1)
      end = signature.length - 1; 
    TypeBinding binding = null;
    if (start == end) {
      BaseTypeBinding baseTypeBinding2;
      VoidTypeBinding voidTypeBinding;
      BaseTypeBinding baseTypeBinding1;
      switch (signature[start]) {
        case 'I':
          baseTypeBinding2 = TypeBinding.INT;
          break;
        case 'Z':
          baseTypeBinding2 = TypeBinding.BOOLEAN;
          break;
        case 'V':
          voidTypeBinding = TypeBinding.VOID;
          break;
        case 'C':
          baseTypeBinding1 = TypeBinding.CHAR;
          break;
        case 'D':
          baseTypeBinding1 = TypeBinding.DOUBLE;
          break;
        case 'B':
          baseTypeBinding1 = TypeBinding.BYTE;
          break;
        case 'F':
          baseTypeBinding1 = TypeBinding.FLOAT;
          break;
        case 'J':
          baseTypeBinding1 = TypeBinding.LONG;
          break;
        case 'S':
          baseTypeBinding1 = TypeBinding.SHORT;
          break;
        default:
          this.problemReporter.corruptedSignature(enclosingType, signature, start);
          break;
      } 
    } else {
      referenceBinding = getTypeFromConstantPoolName(signature, start + 1, end, isParameterized, missingTypeNames);
    } 
    if (isParameterized) {
      if (dimension != 0)
        throw new IllegalStateException(); 
      return (TypeBinding)referenceBinding;
    } 
    if (walker != ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER)
      typeBinding1 = annotateType((TypeBinding)referenceBinding, walker, missingTypeNames); 
    if (dimension != 0)
      arrayBinding = this.typeSystem.getArrayType(typeBinding1, dimension, AnnotatableTypeSystem.flattenedAnnotations(annotationsOnDimensions)); 
    return (TypeBinding)arrayBinding;
  }
  
  private TypeBinding annotateType(TypeBinding binding, ITypeAnnotationWalker walker, char[][][] missingTypeNames) {
    ReferenceBinding referenceBinding;
    TypeBinding typeBinding;
    int depth = binding.depth() + 1;
    if (depth > 1) {
      if (binding.isUnresolvedType())
        referenceBinding = ((UnresolvedReferenceBinding)binding).resolve(this, true); 
      ReferenceBinding referenceBinding1 = referenceBinding;
      depth = 0;
      while (referenceBinding1 != null) {
        depth++;
        if (referenceBinding1.isStatic())
          break; 
        referenceBinding1 = referenceBinding1.enclosingType();
      } 
    } 
    AnnotationBinding[][] annotations = null;
    for (int i = 0; i < depth; i++) {
      AnnotationBinding[] annots = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(((TypeBinding)referenceBinding).id), this, missingTypeNames);
      if (annots != null && annots.length > 0) {
        if (annotations == null)
          annotations = new AnnotationBinding[depth][]; 
        annotations[i] = annots;
      } 
      walker = walker.toNextNestedType();
    } 
    if (annotations != null)
      typeBinding = createAnnotatedType((TypeBinding)referenceBinding, annotations); 
    return typeBinding;
  }
  
  boolean qualifiedNameMatchesSignature(char[][] name, char[] signature) {
    int s = 1;
    for (int i = 0; i < name.length; i++) {
      char[] n = name[i];
      for (int j = 0; j < n.length; j++) {
        if (n[j] != signature[s++])
          return false; 
      } 
      if (signature[s] == ';' && i == name.length - 1)
        return true; 
      if (signature[s++] != '/')
        return false; 
    } 
    return false;
  }
  
  public TypeBinding getTypeFromTypeSignature(SignatureWrapper wrapper, TypeVariableBinding[] staticVariables, ReferenceBinding enclosingType, char[][][] missingTypeNames, ITypeAnnotationWalker walker) {
    int dimension = 0;
    while (wrapper.signature[wrapper.start] == '[') {
      wrapper.start++;
      dimension++;
    } 
    AnnotationBinding[][] annotationsOnDimensions = null;
    if (dimension > 0 && walker != ITypeAnnotationWalker.EMPTY_ANNOTATION_WALKER)
      for (int i = 0; i < dimension; i++) {
        AnnotationBinding[] arrayOfAnnotationBinding = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(0), this, missingTypeNames);
        if (arrayOfAnnotationBinding != Binding.NO_ANNOTATIONS) {
          if (annotationsOnDimensions == null)
            annotationsOnDimensions = new AnnotationBinding[dimension][]; 
          annotationsOnDimensions[i] = arrayOfAnnotationBinding;
        } 
        walker = walker.toNextArrayDimension();
      }  
    if (wrapper.signature[wrapper.start] == 'T') {
      int varStart = wrapper.start + 1;
      int varEnd = wrapper.computeEnd();
      for (int i = staticVariables.length; --i >= 0;) {
        if (CharOperation.equals((staticVariables[i]).sourceName, wrapper.signature, varStart, varEnd))
          return getTypeFromTypeVariable(staticVariables[i], dimension, annotationsOnDimensions, walker, missingTypeNames); 
      } 
      ReferenceBinding initialType = enclosingType;
      while (true) {
        TypeVariableBinding[] enclosingTypeVariables;
        if (enclosingType instanceof BinaryTypeBinding) {
          enclosingTypeVariables = ((BinaryTypeBinding)enclosingType).typeVariables;
        } else {
          enclosingTypeVariables = enclosingType.typeVariables();
        } 
        for (int j = enclosingTypeVariables.length; --j >= 0;) {
          if (CharOperation.equals((enclosingTypeVariables[j]).sourceName, wrapper.signature, varStart, varEnd))
            return getTypeFromTypeVariable(enclosingTypeVariables[j], dimension, annotationsOnDimensions, walker, missingTypeNames); 
        } 
        if ((enclosingType = enclosingType.enclosingType()) == null) {
          this.problemReporter.undefinedTypeVariableSignature(CharOperation.subarray(wrapper.signature, varStart, varEnd), initialType);
          return null;
        } 
      } 
    } 
    boolean isParameterized;
    TypeBinding type = getTypeFromSignature(wrapper.signature, wrapper.start, wrapper.computeEnd(), isParameterized = (wrapper.end == wrapper.bracket), (TypeBinding)enclosingType, missingTypeNames, walker);
    if (!isParameterized)
      return (dimension == 0) ? type : (TypeBinding)createArrayType(type, dimension, AnnotatableTypeSystem.flattenedAnnotations(annotationsOnDimensions)); 
    ReferenceBinding actualType = (ReferenceBinding)type;
    if (actualType instanceof UnresolvedReferenceBinding && 
      CharOperation.indexOf('$', actualType.compoundName[actualType.compoundName.length - 1]) > 0)
      actualType = (ReferenceBinding)BinaryTypeBinding.resolveType((TypeBinding)actualType, this, false); 
    ReferenceBinding actualEnclosing = actualType.enclosingType();
    if (actualEnclosing != null)
      actualEnclosing = (ReferenceBinding)convertToRawType((TypeBinding)actualEnclosing, false); 
    AnnotationBinding[] annotations = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(actualType.id), this, missingTypeNames);
    TypeBinding[] typeArguments = getTypeArgumentsFromSignature(wrapper, staticVariables, enclosingType, actualType, missingTypeNames, walker);
    ParameterizedTypeBinding parameterizedType = createParameterizedType(actualType, typeArguments, actualEnclosing, annotations);
    while (wrapper.signature[wrapper.start] == '.') {
      int memberStart = ++wrapper.start;
      char[] memberName = wrapper.nextWord();
      BinaryTypeBinding.resolveType((TypeBinding)parameterizedType, this, false);
      ReferenceBinding memberType = parameterizedType.genericType().getMemberType(memberName);
      if (memberType == null)
        this.problemReporter.corruptedSignature((TypeBinding)parameterizedType, wrapper.signature, memberStart); 
      walker = walker.toNextNestedType();
      annotations = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(memberType.id), this, missingTypeNames);
      if (wrapper.signature[wrapper.start] == '<') {
        wrapper.start++;
        typeArguments = getTypeArgumentsFromSignature(wrapper, staticVariables, enclosingType, memberType, missingTypeNames, walker);
      } else {
        typeArguments = null;
      } 
      parameterizedType = createParameterizedType(memberType, typeArguments, (ReferenceBinding)parameterizedType, annotations);
    } 
    wrapper.start++;
    return (dimension == 0) ? (TypeBinding)parameterizedType : (TypeBinding)createArrayType((TypeBinding)parameterizedType, dimension, AnnotatableTypeSystem.flattenedAnnotations(annotationsOnDimensions));
  }
  
  private TypeBinding getTypeFromTypeVariable(TypeVariableBinding typeVariableBinding, int dimension, AnnotationBinding[][] annotationsOnDimensions, ITypeAnnotationWalker walker, char[][][] missingTypeNames) {
    AnnotationBinding[] annotations = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(-1), this, missingTypeNames);
    if (annotations != null && annotations != Binding.NO_ANNOTATIONS)
      typeVariableBinding = (TypeVariableBinding)createAnnotatedType((TypeBinding)typeVariableBinding, new AnnotationBinding[][] { annotations }); 
    if (dimension == 0)
      return (TypeBinding)typeVariableBinding; 
    return (TypeBinding)this.typeSystem.getArrayType((TypeBinding)typeVariableBinding, dimension, AnnotatableTypeSystem.flattenedAnnotations(annotationsOnDimensions));
  }
  
  TypeBinding getTypeFromVariantTypeSignature(SignatureWrapper wrapper, TypeVariableBinding[] staticVariables, ReferenceBinding enclosingType, ReferenceBinding genericType, int rank, char[][][] missingTypeNames, ITypeAnnotationWalker walker) {
    TypeBinding bound;
    AnnotationBinding[] annotations;
    switch (wrapper.signature[wrapper.start]) {
      case '-':
        wrapper.start++;
        bound = getTypeFromTypeSignature(wrapper, staticVariables, enclosingType, missingTypeNames, walker.toWildcardBound());
        annotations = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(-1), this, missingTypeNames);
        return (TypeBinding)this.typeSystem.getWildcard(genericType, rank, bound, null, 2, annotations);
      case '+':
        wrapper.start++;
        bound = getTypeFromTypeSignature(wrapper, staticVariables, enclosingType, missingTypeNames, walker.toWildcardBound());
        annotations = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(-1), this, missingTypeNames);
        return (TypeBinding)this.typeSystem.getWildcard(genericType, rank, bound, null, 1, annotations);
      case '*':
        wrapper.start++;
        annotations = BinaryTypeBinding.createAnnotations(walker.getAnnotationsAtCursor(-1), this, missingTypeNames);
        return (TypeBinding)this.typeSystem.getWildcard(genericType, rank, null, null, 0, annotations);
    } 
    return getTypeFromTypeSignature(wrapper, staticVariables, enclosingType, missingTypeNames, walker);
  }
  
  boolean isMissingType(char[] typeName) {
    for (int i = (this.missingTypes == null) ? 0 : this.missingTypes.size(); --i >= 0; ) {
      MissingTypeBinding missingType = this.missingTypes.get(i);
      if (CharOperation.equals(missingType.sourceName, typeName))
        return true; 
    } 
    return false;
  }
  
  boolean isPackage(char[][] compoundName, char[] name) {
    if (compoundName == null || compoundName.length == 0)
      return this.nameEnvironment.isPackage(null, name); 
    return this.nameEnvironment.isPackage(compoundName, name);
  }
  
  public MethodVerifier methodVerifier() {
    if (this.verifier == null)
      this.verifier = newMethodVerifier(); 
    return this.verifier;
  }
  
  public MethodVerifier newMethodVerifier() {
    return (MethodVerifier)new MethodVerifier15(this);
  }
  
  public void releaseClassFiles(ClassFile[] classFiles) {
    for (int i = 0, fileCount = classFiles.length; i < fileCount; i++)
      this.classFilePool.release(classFiles[i]); 
  }
  
  public void reset() {
    this.defaultPackage = new PackageBinding(this);
    this.defaultImports = null;
    this.knownPackages = new HashtableOfPackage();
    this.accessRestrictions = new HashMap<>(3);
    this.verifier = null;
    this.uniqueParameterizedGenericMethodBindings = new SimpleLookupTable(3);
    this.uniquePolymorphicMethodBindings = new SimpleLookupTable(3);
    this.uniqueGetClassMethodBinding = null;
    this.missingTypes = null;
    this.typesBeingConnected = new HashSet<>();
    for (int i = this.units.length; --i >= 0;)
      this.units[i] = null; 
    this.lastUnitIndex = -1;
    this.lastCompletedUnitIndex = -1;
    this.unitBeingCompleted = null;
    this.classFilePool.reset();
    this.typeSystem.reset();
  }
  
  public void setAccessRestriction(ReferenceBinding type, AccessRestriction accessRestriction) {
    if (accessRestriction == null)
      return; 
    type.modifiers |= 0x40000;
    this.accessRestrictions.put(type, accessRestriction);
  }
  
  void updateCaches(UnresolvedReferenceBinding unresolvedType, ReferenceBinding resolvedType) {
    this.typeSystem.updateCaches(unresolvedType, resolvedType);
  }
  
  public LookupEnvironment(ITypeRequestor typeRequestor, CompilerOptions globalOptions, ProblemReporter problemReporter, INameEnvironment nameEnvironment) {
    this.resolutionListeners = new IQualifiedTypeResolutionListener[0];
    this.typeRequestor = typeRequestor;
    this.globalOptions = globalOptions;
    this.problemReporter = problemReporter;
    this.defaultPackage = new PackageBinding(this);
    this.defaultImports = null;
    this.nameEnvironment = nameEnvironment;
    this.knownPackages = new HashtableOfPackage();
    this.uniqueParameterizedGenericMethodBindings = new SimpleLookupTable(3);
    this.uniquePolymorphicMethodBindings = new SimpleLookupTable(3);
    this.missingTypes = null;
    this.accessRestrictions = new HashMap<>(3);
    this.classFilePool = ClassFilePool.newInstance();
    this.typesBeingConnected = new HashSet<>();
    this.typeSystem = (this.globalOptions.sourceLevel >= 3407872L && this.globalOptions.storeAnnotations) ? (TypeSystem)new AnnotatableTypeSystem(this) : new TypeSystem(this);
  }
  
  public void addResolutionListener(IQualifiedTypeResolutionListener resolutionListener) {
    int length = this.resolutionListeners.length;
    for (int i = 0; i < length; i++) {
      if (this.resolutionListeners[i].equals(resolutionListener))
        return; 
    } 
    System.arraycopy(this.resolutionListeners, 0, 
        this.resolutionListeners = new IQualifiedTypeResolutionListener[length + 1], 0, length);
    this.resolutionListeners[length] = resolutionListener;
  }
  
  public TypeBinding getUnannotatedType(TypeBinding typeBinding) {
    return this.typeSystem.getUnannotatedType(typeBinding);
  }
  
  public TypeBinding[] getAnnotatedTypes(TypeBinding type) {
    return this.typeSystem.getAnnotatedTypes(type);
  }
  
  public AnnotationBinding[] filterNullTypeAnnotations(AnnotationBinding[] typeAnnotations) {
    if (typeAnnotations.length == 0)
      return typeAnnotations; 
    AnnotationBinding[] filtered = new AnnotationBinding[typeAnnotations.length];
    int count = 0;
    for (int i = 0; i < typeAnnotations.length; i++) {
      AnnotationBinding typeAnnotation = typeAnnotations[i];
      if (typeAnnotation == null) {
        count++;
      } else if (!typeAnnotation.type.hasNullBit(96)) {
        filtered[count++] = typeAnnotation;
      } 
    } 
    if (count == 0)
      return Binding.NO_ANNOTATIONS; 
    if (count == typeAnnotations.length)
      return typeAnnotations; 
    System.arraycopy(filtered, 0, filtered = new AnnotationBinding[count], 0, count);
    return filtered;
  }
  
  public boolean containsNullTypeAnnotation(IBinaryAnnotation[] typeAnnotations) {
    if (typeAnnotations.length == 0)
      return false; 
    for (int i = 0; i < typeAnnotations.length; i++) {
      IBinaryAnnotation typeAnnotation = typeAnnotations[i];
      char[] typeName = typeAnnotation.getTypeName();
      if (typeName != null && typeName.length >= 3 && typeName[0] == 'L') {
        char[][] name = CharOperation.splitOn('/', typeName, 1, typeName.length - 1);
        if (getNullAnnotationBit(name) != 0)
          return true; 
      } 
    } 
    return false;
  }
  
  public boolean containsNullTypeAnnotation(AnnotationBinding[] typeAnnotations) {
    if (typeAnnotations.length == 0)
      return false; 
    for (int i = 0; i < typeAnnotations.length; i++) {
      AnnotationBinding typeAnnotation = typeAnnotations[i];
      if (typeAnnotation.type.hasNullBit(96))
        return true; 
    } 
    return false;
  }
}
